from typing import ClassVar

class IcfSolverParameters:
    __fields__: ClassVar[tuple] = ...  # read-only
    alpha_max: float
    enable_hessian_reuse: bool
    hessian_reuse_target_iterations: int
    linesearch_tolerance: float
    max_iterations: int
    max_linesearch_iterations: int
    min_tolerance: float
    print_solver_stats: bool
    use_dense_algebra: bool
    def __init__(self, **kwargs) -> None: ...
    def __copy__(self) -> IcfSolverParameters: ...
    def __deepcopy__(self, arg0: dict) -> IcfSolverParameters: ...



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_iiwa_command(object):
    """ Commands a single set of joint states for the arm. """

    __slots__ = ["utime", "num_joints", "joint_position", "num_torques", "joint_torque"]

    __typenames__ = ["int64_t", "int32_t", "double", "int32_t", "double"]

    __dimensions__ = [None, None, ["num_joints"], None, ["num_torques"]]

    def __init__(self):
        self.utime = 0
        """
        The timestamp in microseconds.
        LCM Type: int64_t
        """

        self.num_joints = 0
        """
        The reference joint positions. They must be sent when the arm is in
        position control mode, but must be of size zero in torque control mode.
        LCM Type: int32_t
        """

        self.joint_position = []
        """ LCM Type: double[num_joints] """
        self.num_torques = 0
        """
        TODO(eric.cousineau): The above name should really be num_position(s).
        However, this may be moot if we transition to more homogeneous middleware
        message types and translation layers.
        The reference joint torques. They should only be sent when the arm is in
        torque control mode. When only positions are being sent, num_torques
        should be set to zero.
        LCM Type: int32_t
        """

        self.joint_torque = []
        """ LCM Type: double[num_torques] """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_iiwa_command._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">qi", self.utime, self.num_joints))
        buf.write(struct.pack('>%dd' % self.num_joints, *self.joint_position[:self.num_joints]))
        buf.write(struct.pack(">i", self.num_torques))
        buf.write(struct.pack('>%dd' % self.num_torques, *self.joint_torque[:self.num_torques]))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_iiwa_command._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_iiwa_command._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_iiwa_command()
        self.utime, self.num_joints = struct.unpack(">qi", buf.read(12))
        self.joint_position = struct.unpack('>%dd' % self.num_joints, buf.read(self.num_joints * 8))
        self.num_torques = struct.unpack(">i", buf.read(4))[0]
        self.joint_torque = struct.unpack('>%dd' % self.num_torques, buf.read(self.num_torques * 8))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_iiwa_command in parents: return 0
        tmphash = (0x6ee3e3b9c640a99a) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_iiwa_command._packed_fingerprint is None:
            lcmt_iiwa_command._packed_fingerprint = struct.pack(">Q", lcmt_iiwa_command._get_hash_recursive([]))
        return lcmt_iiwa_command._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_iiwa_command._get_packed_fingerprint())[0]




import flags
import numpy
import pydrake.systems.framework
from pydrake.common.cpp_template import LuenbergerObserver_ as LuenbergerObserver_
from typing import ClassVar

class LuenbergerObserver(pydrake.systems.framework.LeafSystem):
    def __init__(self, observed_system: pydrake.systems.framework.System, observed_system_context: pydrake.systems.framework.Context, observer_gain: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> None: ...
    def L(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort: ...
    def observer_gain(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

class LuenbergerObserver_𝓣AutoDiffXd𝓤(pydrake.systems.framework.LeafSystem_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TAutoDiffXdU, observed_system_context: pydrake.systems.framework.Context_TAutoDiffXdU, observer_gain: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> None: ...
    def L(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TAutoDiffXdU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def observer_gain(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

class LuenbergerObserver_𝓣Expression𝓤(pydrake.systems.framework.LeafSystem_TExpressionU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TExpressionU, observed_system_context: pydrake.systems.framework.Context_TExpressionU, observer_gain: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> None: ...
    def L(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TExpressionU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def observer_gain(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

class _TemporaryName_N5drake7systems10estimators18LuenbergerObserverINS_2ad8AutoDiffEEE(pydrake.systems.framework.LeafSystem_TAutoDiffXdU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TAutoDiffXdU, observed_system_context: pydrake.systems.framework.Context_TAutoDiffXdU, observer_gain: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> None: ...
    def L(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TAutoDiffXdU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TAutoDiffXdU: ...
    def observer_gain(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

class _TemporaryName_N5drake7systems10estimators18LuenbergerObserverINS_8symbolic10ExpressionEEE(pydrake.systems.framework.LeafSystem_TExpressionU):
    _original_name: ClassVar[str] = ...
    _original_qualname: ClassVar[str] = ...
    def __init__(self, observed_system: pydrake.systems.framework.System_TExpressionU, observed_system_context: pydrake.systems.framework.Context_TExpressionU, observer_gain: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> None: ...
    def L(self) -> numpy.ndarray[numpy.float64[m, n]]: ...
    def get_estimated_state_output_port(self) -> pydrake.systems.framework.OutputPort_TExpressionU: ...
    def get_observed_system_input_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def get_observed_system_output_input_port(self) -> pydrake.systems.framework.InputPort_TExpressionU: ...
    def observer_gain(self) -> numpy.ndarray[numpy.float64[m, n]]: ...

def DiscreteTimeSteadyStateKalmanFilter(A: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], C: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], W: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], V: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> numpy.ndarray[numpy.float64[m, n]]: ...
def SteadyStateKalmanFilter(A: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], C: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], W: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous], V: numpy.ndarray[numpy.float64[m, n], flags.f_contiguous]) -> numpy.ndarray[numpy.float64[m, n]]: ...



"""This file contains the implementation of to_sympy and from_sympy as used by
`_symbolic_extra.py`. It is loaded is loaded lazily (on demand), so that Drake
does not directly depend on SymPy.
"""

import math
import operator

import sympy
from sympy.printing.pycode import MpmathPrinter

import pydrake.symbolic
from pydrake.symbolic import (
    Expression,
    ExpressionKind,
    Formula,
    FormulaKind,
    Variable,
)


def _no_change(x):
    """This function simply returns its argument. It is used below in the table
    of constructors to signify that no type-conversion is required.
    """
    return x


def _handle_constant(x):
    """This function is used to cast constant expressions to the appropriate
    sympy type.
    """
    if not math.isinf(x) and not math.isnan(x) and int(x) == x:
        return sympy.Integer(x)
    return sympy.Float(x)


def _make_sympy_if_then_else(
    cond: sympy.Expr, expr_then: sympy.Expr, expr_else: sympy.Expr
):
    """Returns the SymPy spelling of `{then_expr} if {cond} else {expr_else}`,
    also known as the "ternary conditional" operator.
    """
    # N.B. We can't use sympy.ITE -- that operates on three booleans; the Drake
    # if_then_else operates on (bool, float, float).
    return sympy.Piecewise((expr_then, cond), (expr_else, True))


# This table maps from ExpressionKind and FormulaKind to the SymPy constructor
# for that kind of term; this forms the core of `pydrake.symbolic.to_sympy()`.
_SYMPY_CONSTRUCTOR = {
    ExpressionKind.Constant: _handle_constant,
    # Like Drake's NaN, the SymPy NaN also rejects, e.g., comparison operators.
    # That's more like what we want than `math.nan`, which allows comparison.
    ExpressionKind.NaN: lambda _: sympy.nan,
    # Use the SymPy preferred spelling of bools.
    FormulaKind.True_: lambda: sympy.true,
    FormulaKind.False_: lambda: sympy.false,
    # SymPy doesn't need any extra call to convert a variable to an expression.
    ExpressionKind.Var: _no_change,
    FormulaKind.Var: _no_change,
    # The rest is all boring stuff, in alphabetical order.
    ExpressionKind.Abs: sympy.Abs,
    ExpressionKind.Acos: sympy.acos,
    ExpressionKind.Add: sympy.Add,
    ExpressionKind.Asin: sympy.asin,
    ExpressionKind.Atan: sympy.atan,
    ExpressionKind.Atan2: sympy.atan2,
    ExpressionKind.Ceil: sympy.ceiling,
    ExpressionKind.Cos: sympy.cos,
    ExpressionKind.Cosh: sympy.cosh,
    ExpressionKind.Div: operator.truediv,
    ExpressionKind.Exp: sympy.exp,
    ExpressionKind.Floor: sympy.floor,
    ExpressionKind.IfThenElse: _make_sympy_if_then_else,
    ExpressionKind.Log: sympy.log,
    ExpressionKind.Max: sympy.Max,
    ExpressionKind.Min: sympy.Min,
    ExpressionKind.Mul: sympy.Mul,
    ExpressionKind.Pow: sympy.Pow,
    ExpressionKind.Sin: sympy.sin,
    ExpressionKind.Sinh: sympy.sinh,
    ExpressionKind.Sqrt: sympy.sqrt,
    ExpressionKind.Tan: sympy.tan,
    ExpressionKind.Tanh: sympy.tanh,
    FormulaKind.And: sympy.And,
    FormulaKind.Eq: sympy.Equality,
    FormulaKind.Geq: sympy.GreaterThan,
    FormulaKind.Gt: sympy.StrictGreaterThan,
    FormulaKind.Leq: sympy.LessThan,
    FormulaKind.Lt: sympy.StrictLessThan,
    FormulaKind.Neq: sympy.Unequality,
    FormulaKind.Not: sympy.Not,
    FormulaKind.Or: sympy.Or,
    # We don't support converting these kinds of terms (yet).
    ExpressionKind.UninterpretedFunction: None,
    FormulaKind.Forall: None,
    FormulaKind.Isnan: None,
    FormulaKind.PositiveSemidefinite: None,
}


def _var_to_sympy(drake_var: Variable, *, memo: dict):
    """Converts a Drake variable into a SymPy Variable.

    If the Drake variable was already in the `memo` dict, returns the value in
    the dict. Otherwise, creates and returns a new sympy.Dummy() variable.
    """
    sympy_var = memo.get(drake_var.get_id())
    if sympy_var is None:
        drake_type = drake_var.get_type()  # noqa: F841 (unused-variable)
        assumptions = {
            # TODO(jwnimmer-tri) Use drake_type to fill in the assumptions.
        }
        sympy_var = sympy.Dummy(
            name=drake_var.get_name(),
            dummy_index=drake_var.get_id(),
            **assumptions,
        )
        memo[drake_var.get_id()] = sympy_var
        memo[sympy_var] = drake_var
    return sympy_var


def _var_from_sympy(sympy_var: sympy.Dummy, *, memo: dict):
    """Converts a SymPy variable into a Drake Variable.

    At the moment, the *only* supported conversion is to look up the variable
    in `memo`, i.e., the user may pre-define the mapping.

    TODO(jwnimmer-tri) This could probably have a better fallback plan.
    """
    drake_var = memo.get(sympy_var)
    if drake_var is None:
        raise NotImplementedError(
            f"The SymPy variable {sympy_var} is missing from the `memo` dict."
        )
    return drake_var


def _to_sympy(
    x: float | int | bool | Variable | Expression | Formula,
    *,
    memo: dict = None,
) -> float | int | bool | sympy.Expr:
    """This is the private implementation of pydrake.symbolic.to_sympy().
    Refer to that module-level function for the full docstring.

    TODO(jwnimmer-tri) Also support Polynomial, Monomial, etc.
    """
    if isinstance(x, (float, int, bool)):
        return x
    if isinstance(x, Variable):
        return _var_to_sympy(drake_var=x, memo=memo)
    try:
        kind = x.get_kind()
    except AttributeError:
        kind = None
    if kind is None:
        raise NotImplementedError(
            f"Cannot create a SymPy object from the given object {x!r}"
        )
    sympy_constructor = _SYMPY_CONSTRUCTOR.get(kind)
    if sympy_constructor is None:
        raise NotImplementedError(
            f"Cannot create a SymPy object from "
            f"the given pydrake {kind} object {x!r}"
        )
    _, drake_args = x.Unapply()
    sympy_args = [_to_sympy(arg, memo=memo) for arg in drake_args]
    return sympy_constructor(*sympy_args)


class _DrakePrinter(MpmathPrinter):
    """A slightly customized Printer class that handles a few unique spellings
    necessary for writing Drake code.
    """

    def __init__(self):
        super().__init__(
            {
                "fully_qualified_modules": False,
                "inline": True,
                "allow_unknown_functions": True,
                "user_functions": {},
            }
        )

    def _print_drake_logical_op(self, expr, op):
        """Uses pydrake.symbolic.logical_{op} instead of the Python built-in
        boolean operators (`and`, `or`, and `not`).
        """
        args = [self._print(arg) for arg in expr.args]
        return f"logical_{op}({', '.join(args)})"

    def _print_And(self, expr):
        return self._print_drake_logical_op(expr, "and")

    def _print_Or(self, expr):
        return self._print_drake_logical_op(expr, "or")

    def _print_Not(self, expr):
        return self._print_drake_logical_op(expr, "not")

    def _print_Piecewise(self, expr):
        """Undo the effect of _make_sympy_if_then_else, by converting a list of
        `((cond1, expr1), (cond2, expr2), ...)` pairs into a nested chain of
        `if_then_else(cond1, expr1, if_then_else(cond2, expr2, ...))` calls.
        """
        assert len(expr.args) > 0
        if expr.args[-1].cond not in (True, sympy.true):
            raise NotImplementedError(
                "Piecewise functions must always have a value; the final "
                "condition must be the literal value `True`."
            )
        result = [self._print(expr.args[-1].expr)]
        for arg in reversed(expr.args[:-1]):
            arg_cond = self._print(arg.cond)
            arg_expr = self._print(arg.expr)
            result.insert(0, f"if_then_else({arg_cond}, {arg_expr}, ")
            result.append(")")
        return "".join(result)


_DRAKE_PRINTER = _DrakePrinter()


def _lambdify(*, expr, args):
    """Wraps sympy.lambdify with extra hard-coded arguments."""
    return sympy.lambdify(
        expr=expr,
        args=args,
        modules=pydrake.symbolic,
        printer=_DRAKE_PRINTER,
        use_imps=False,
        docstring_limit=0,
    )


def _from_sympy(
    x: float | int | bool | sympy.Expr, *, memo: dict = None
) -> float | int | bool | Variable | Expression | Formula:
    """This is the private implementation of pydrake.symbolic.from_sympy().
    Refer to that module-level function for the full docstring.
    """
    # Return non-SymPy inputs as-is.
    if isinstance(x, (float, int, bool)):
        return x
    # Return constants quickly.
    if x.is_number:
        return float(x.evalf())
    # Find the SymPy variables in `x` and look up the matching Drake variables.
    sympy_vars = []
    drake_vars = []
    for item in x.atoms():
        if item.is_number:
            continue
        if isinstance(item, sympy.logic.boolalg.BooleanAtom):
            continue
        if isinstance(item, sympy.Dummy):
            sympy_vars.append(item)
            drake_vars.append(_var_from_sympy(item, memo=memo))
            continue
        raise NotImplementedError(f"Unsupported atom {item!r} ({type(item)})")
    # Convert the SymPy expression to a Python function of the variables.
    drake_func = _lambdify(expr=x, args=sympy_vars)
    # Call the Python function using the Drake variables; this returns the
    # Drake symbolic expression.
    return drake_func(*drake_vars)



"""LCM type definitions
This file automatically generated by lcm.
DO NOT MODIFY BY HAND!!!!
"""


from io import BytesIO
import struct

class lcmt_allegro_status(object):

    __slots__ = ["utime", "num_joints", "joint_position_measured", "joint_velocity_estimated", "joint_position_commanded", "joint_torque_commanded"]

    __typenames__ = ["int64_t", "int32_t", "double", "double", "double", "double"]

    __dimensions__ = [None, None, ["num_joints"], ["num_joints"], ["num_joints"], ["num_joints"]]

    def __init__(self):
        self.utime = 0
        """
        The timestamp in microseconds.
        LCM Type: int64_t
        """

        self.num_joints = 0
        """ LCM Type: int32_t """
        self.joint_position_measured = []
        """ LCM Type: double[num_joints] """
        self.joint_velocity_estimated = []
        """ LCM Type: double[num_joints] """
        self.joint_position_commanded = []
        """ LCM Type: double[num_joints] """
        self.joint_torque_commanded = []
        """ LCM Type: double[num_joints] """

    def encode(self):
        buf = BytesIO()
        buf.write(lcmt_allegro_status._get_packed_fingerprint())
        self._encode_one(buf)
        return buf.getvalue()

    def _encode_one(self, buf):
        buf.write(struct.pack(">qi", self.utime, self.num_joints))
        buf.write(struct.pack('>%dd' % self.num_joints, *self.joint_position_measured[:self.num_joints]))
        buf.write(struct.pack('>%dd' % self.num_joints, *self.joint_velocity_estimated[:self.num_joints]))
        buf.write(struct.pack('>%dd' % self.num_joints, *self.joint_position_commanded[:self.num_joints]))
        buf.write(struct.pack('>%dd' % self.num_joints, *self.joint_torque_commanded[:self.num_joints]))

    @staticmethod
    def decode(data: bytes):
        if hasattr(data, 'read'):
            buf = data
        else:
            buf = BytesIO(data)
        if buf.read(8) != lcmt_allegro_status._get_packed_fingerprint():
            raise ValueError("Decode error")
        return lcmt_allegro_status._decode_one(buf)

    @staticmethod
    def _decode_one(buf):
        self = lcmt_allegro_status()
        self.utime, self.num_joints = struct.unpack(">qi", buf.read(12))
        self.joint_position_measured = struct.unpack('>%dd' % self.num_joints, buf.read(self.num_joints * 8))
        self.joint_velocity_estimated = struct.unpack('>%dd' % self.num_joints, buf.read(self.num_joints * 8))
        self.joint_position_commanded = struct.unpack('>%dd' % self.num_joints, buf.read(self.num_joints * 8))
        self.joint_torque_commanded = struct.unpack('>%dd' % self.num_joints, buf.read(self.num_joints * 8))
        return self

    @staticmethod
    def _get_hash_recursive(parents):
        if lcmt_allegro_status in parents: return 0
        tmphash = (0x8d235a54b2c7b3d1) & 0xffffffffffffffff
        tmphash  = (((tmphash<<1)&0xffffffffffffffff) + (tmphash>>63)) & 0xffffffffffffffff
        return tmphash
    _packed_fingerprint = None

    @staticmethod
    def _get_packed_fingerprint():
        if lcmt_allegro_status._packed_fingerprint is None:
            lcmt_allegro_status._packed_fingerprint = struct.pack(">Q", lcmt_allegro_status._get_hash_recursive([]))
        return lcmt_allegro_status._packed_fingerprint

    def get_hash(self):
        """Get the LCM hash of the struct"""
        return struct.unpack(">Q", lcmt_allegro_status._get_packed_fingerprint())[0]




import logging
from pathlib import Path
import xml.etree.ElementTree

import numpy as np

from pydrake.geometry import (
    ReadObjToTriangleSurfaceMesh,
)
from pydrake.multibody.tree import (
    CalcSpatialInertia,
    SpatialInertia,
)

_logger = logging.getLogger("drake")

_PACKAGE_TEMPLATE = """
<?xml version="1.0"?>
<package format="2">
  <name>{name}</name>
  <version>0.0.0</version>
  <description>
    Package generated by mesh_to_model from mesh file: {mesh_name}.
  </description>
  <maintainer email="nobody@example.com">Nobody</maintainer>
  <license>unknown</license>
</package>
"""

_SDF_TEMPLATE = """<?xml version='1.0'?>
<sdf xmlns:drake="http://drake.mit.edu" version="1.8">
  <model name='{name}'>
    <link name='{name}'>
      <inertial>
        <pose>{inertial_pose}</pose>
        <mass>{mass}</mass>
        <inertia>
          <ixx>{ixx}</ixx>
          <ixy>{ixy}</ixy>
          <ixz>{ixz}</ixz>
          <iyy>{iyy}</iyy>
          <iyz>{iyz}</iyz>
          <izz>{izz}</izz>
        </inertia>
      </inertial>
      <visual name='visual'>
        <pose>{geometry_position} 0 0 0</pose>
        <geometry>
          <mesh>
            <uri>{mesh_uri}</uri>
            <scale>{mesh_scale}</scale>
          </mesh>
        </geometry>
      </visual>
      <collision name='collision'>
        <pose>{geometry_position} 0 0 0</pose>
        <geometry>
          <mesh>
            <uri>{mesh_uri}</uri>
            <scale>{mesh_scale}</scale>
          </mesh>
        </geometry>
        <!-- The proximity properties immediately below enable the hydroelastic
             contact model for this body. To use point contact instead, delete
             or comment out the <drake:rigid_hydroelastic> element. -->
        <drake:proximity_properties>
          <drake:mu_dynamic>0.5</drake:mu_dynamic>
          <drake:mu_static>0.5</drake:mu_static>
          <drake:rigid_hydroelastic/>
        </drake:proximity_properties>
      </collision>
    </link>
  </model>
</sdf>
"""


class MeshModelMaker:
    """Converts a mesh file into a model, documenting the work as it goes."""

    def __init__(
        self,
        *,
        mesh_path: Path,
        output_dir: Path = None,
        scale: float = 1.0,
        model_name: str = None,
        density: float = 1000.0,
        mass: float = None,
        at_com: bool = False,
        p_GoBo: np.array = None,
        encoded_package: str = "none",
    ):
        """
        Read the given mesh definition and construct a corresponding SDFormat
        file representing a single model with a single body. The body's mass
        properties are inferred from the volume enclosed by the mesh.

        For more details, read the documentation in mesh_to_model.py.

        Either ``density`` or ``mass`` defines how the mass is specified. At
        least one must be provided. If mass is provided, it will be used and
        density will be ignored.

        ``at_com`` and ``p_GoBo`` control the relative position of the
        geometry and body frames. There are three options:
        - Body origin at geometry origin (at_com=False, p_GoBo=None).
        - Body origin at geometry center of mass (at_com=True, p_GoBo=None).
        - Body origin at arbitrary point (measured and expressed in geometry
          frame) (at_com=False, p_GoBo=[x_cm, y_cm, z_cm]).
        It is an error to define at_com=True and p_GoBo not None.

        If ``output_dir`` is not None, it must name a valid directory.
        """
        if density is None and mass is None:
            raise ValueError("Neither mass nor density were provided.")
        if density is not None and mass is not None:
            density = None
        if at_com and p_GoBo is not None:
            raise ValueError("Specify either at_com=True or p_GoBo, not both.")
        if output_dir is None:
            output_dir = None if mesh_path is None else mesh_path.parent
        elif not output_dir.is_dir():
            raise ValueError(
                "The output directory must name an existing "
                f"directory; given {output_dir}."
            )
        self.mesh_path = mesh_path
        self.output_dir = output_dir
        self.scale = scale
        self.model_name = model_name
        self.density = density
        self.mass = mass
        self.at_com = at_com
        self.p_GoBo = p_GoBo
        self.encoded_package = encoded_package

    def make_model(self):
        """
        Creates a model from the mesh given at mesh_path and writes the
        corresponding model to the file at model_path based on the maker's
        configurations.

        See the documentation for mesh_to_model.py for a discussion of the
        configuration parameters.
        """
        _logger.info(f"Creating a model from: {self.mesh_path}")

        # Figure out the mesh URI and model names.
        mesh_uri = MeshModelMaker._make_mesh_uri(
            self.encoded_package, self.mesh_path
        )

        if self.scale <= 0:
            raise ValueError(
                f"Scale value must be positive, given {self.scale}."
            )

        mesh_G = ReadObjToTriangleSurfaceMesh(
            filename=str(self.mesh_path), scale=self.scale
        )
        p_GoMin, p_GoMax = mesh_G.CalcBoundingBox()
        size = p_GoMax - p_GoMin
        _logger.info("Mesh-model summary:")
        _logger.info(
            f"    Bounding box (in geometry frame): {size[0]} x "
            f"{size[1]} x {size[2]} meters"
        )

        # TODO(SeanCurtis-TRI): Confirm that the mesh is watertight.

        if self.mass is not None:
            # If mass is defined, it wins.
            # Neither the unit inertia nor the center of mass depend on the
            # density. So, we can compute those quantities for an arbitrary
            # density and then reconfigure the SpatialInertia with the desired
            # mass.
            density = 1.0
            M_GGo_G = CalcSpatialInertia(mesh=mesh_G, density=density)
            volume = M_GGo_G.get_mass() / density
            M_GGo_G = SpatialInertia(
                mass=self.mass,
                p_PScm_E=M_GGo_G.get_com(),
                G_SP_E=M_GGo_G.get_unit_inertia(),
            )
        else:
            M_GGo_G = CalcSpatialInertia(mesh=mesh_G, density=self.density)
            volume = M_GGo_G.get_mass() / self.density

        p_GoGcm = M_GGo_G.get_com()
        M_GGcm_G = M_GGo_G.Shift(p_GoGcm)
        mass = M_GGo_G.get_mass()

        _logger.info(f"    Mass: {mass} kg")

        if self.p_GoBo is None:
            if self.at_com:
                p_BoGo = -p_GoGcm
                p_BoBcm = (0.0, 0.0, 0.0)
            else:
                # Default case.
                p_BoGo = (0.0, 0.0, 0.0)
                p_BoBcm = p_GoGcm
        else:
            p_BoGo = -self.p_GoBo
            p_BoBcm = p_GoGcm - self.p_GoBo

        _logger.info(f"    p_GoGcm: [{p_GoGcm[0]}, {p_GoGcm[1]}, {p_GoGcm[2]}]")
        _logger.info(f"    p_BoBcm: [{p_BoBcm[0]}, {p_BoBcm[1]}, {p_BoBcm[2]}]")

        # The Empire State building conveniently exemplifies 1e6 cubic meters.
        # https://en.wikipedia.org/wiki/Empire_State_Building#Interior
        huge_volume = 1e6
        if volume > huge_volume:
            _logger.warning(
                f"Mesh volume [{volume} m³] exceeds the volume"
                f" of the Empire State Building"
                f" [{huge_volume} m³]. Consider using the"
                f" --scale option."
            )

        # Medium sand tops out at 0.5 mm diameter, leading to the approximate
        # volume shown below.
        # https://en.wikipedia.org/wiki/Grain_size
        tiny_volume = 6e-11
        if volume < tiny_volume:
            _logger.warning(
                f"Mesh volume [{volume} m³] is smaller than the"
                f" volume of a grain of sand [{tiny_volume} m³]."
                f" Consider using the --scale option."
            )

        # In SDF files, the inertia tensor is always reported around the center
        # of mass.
        I_GGcm_G = M_GGcm_G.CalcRotationalInertia()
        moments = I_GGcm_G.get_moments()
        products = I_GGcm_G.get_products()

        subs = dict()
        # model_name is None *or* model_name == "" prefers mesh_stem.
        subs["name"] = self.model_name or self.mesh_path.stem
        # With zero rotation, this is simply the position of com in B.
        p_BoBcm_str = " ".join([str(x) for x in p_BoBcm])
        subs["inertial_pose"] = f"{p_BoBcm_str} 0 0 0"
        subs["mass"] = mass
        subs["ixx"] = moments[0]
        subs["iyy"] = moments[1]
        subs["izz"] = moments[2]
        subs["ixy"] = products[0]
        subs["ixz"] = products[1]
        subs["iyz"] = products[2]
        subs["mesh_uri"] = mesh_uri
        subs["mesh_scale"] = "{s} {s} {s}".format(s=self.scale)
        subs["geometry_position"] = " ".join([str(x) for x in p_BoGo])

        output_path = self.output_dir / f"{self.mesh_path.stem}.sdf"
        with open(output_path, "w") as f:
            f.write(_SDF_TEMPLATE.format(**subs))

    @staticmethod
    def _make_mesh_uri(package_spec: str, mesh_path: Path):
        """
        Given an encoded package protocol and path to the mesh, creates the
        uri to be used. May create a package.xml file in the same directory as
        the mesh file (depending on the value of package_spec).

        If package_spec contains the path to a package.xml file, that file must
        be located in an ancestor directory of the mesh file.

        Args:
            package_spec: Can be one of "none", "auto", or a path to a
                package.xml file.
            mesh_path: The path to the mesh file being processed. Used if
                package_spec is not "none". A package.xml file in that
                directory will be used or created.

        Returns:
            A string representing the uri to use the SDFormat file to reference
            the mesh.
        """
        mesh_name = mesh_path.name

        if package_spec == "none":
            return mesh_name

        package_path = None
        if package_spec.endswith("package.xml"):
            # The mesh must lie in the tree rooted at the package.
            package_spec = Path(package_spec)
            if not package_spec.exists():
                raise ValueError(
                    "The indicated package.xml cannot be found: "
                    f"'{package_spec}'"
                )
            abs_package = package_spec.resolve().absolute()
            abs_mesh = mesh_path.resolve().absolute()
            try:
                mesh_rel_path = abs_mesh.relative_to(abs_package.parent)
            except ValueError:
                raise ValueError(
                    "When specifying the package, the mesh must be located in "
                    f"the file tree of the package.\n  package: {abs_package}"
                    f"\n  mesh: {abs_mesh}"
                )
            package_name = MeshModelMaker._read_package_name(package_spec)
            package_path = f"package://{package_name}/{mesh_rel_path}"

        if package_spec == "auto":
            mesh_dir = mesh_path.resolve().absolute().parent
            candidate_package_path = mesh_dir / "package.xml"
            if candidate_package_path.exists():
                package_name = MeshModelMaker._read_package_name(
                    candidate_package_path
                )
            else:
                # The name of the *package* is the name of the mesh file's
                # directory.
                package_name = mesh_dir.name
                subs = {"name": package_name, "mesh_name": mesh_name}
                with open(candidate_package_path, "w") as f:
                    f.write(_PACKAGE_TEMPLATE.format(**subs))
            package_path = f"package://{package_name}/{mesh_name}"

        if package_path is not None:
            return package_path

        raise ValueError(
            f"Unrecognized package specification: '{package_spec}'."
        )

    @staticmethod
    def _read_package_name(package_path: Path):
        """
        Reads the ROS package manifest file at the given path, returning the
        value of the name tag.
        """
        root = xml.etree.ElementTree.parse(package_path).getroot()
        name_element = root.find("name")
        if name_element is None:
            raise ValueError(
                "The provided package.xml is missing a 'name' "
                f"element: {package_path}."
            )
        return name_element.text.strip()



"""Provides containers for tracking instantiations of C++ templates."""

import inspect
import types

from pydrake import _is_building_documentation
from pydrake.common import _MangledName, pretty_class_name
from pydrake.common.cpp_param import get_param_canonical, get_param_names
from pydrake.common.deprecation import _warn_deprecated


def _get_module_from_stack(frame=2):
    # Infers module name from call stack.
    return inspect.getmodule(inspect.stack()[frame][0])


def _is_pybind11_type_error(e):
    return any(
        [
            "incompatible function arguments" in str(e),
            "incompatible constructor arguments" in str(e),
        ]
    )


def get_or_init(scope, name, template_cls, *args, **kwargs):
    """Gets an existing template from a scope if it exists; otherwise, it will
    be created and registered.

    If `scope` does not have the attribute `name`, then it will be created as
    `template_cls(name, *args, **kwargs)`.
    (`module_name=...` is also set, but should not be important).

    Args:
        scope: Scope that contains the template object (this may not
            necessarily be the scope of the instantiation).
        name: Name of the template object.
        template_cls: Class (either `TemplateBase` or a derivative).
        args, kwargs: Passed to the template class's constructor.

    Returns:
        The existing (or newly created) template object.
    """
    template = getattr(scope, name, None)
    if template is None:
        template = template_cls(name, *args, scope=scope, **kwargs)
        setattr(scope, name, template)
    return template


class _Deprecation:
    # Denotes a (message, date) tuple.
    def __init__(self, *, message, date):
        self.message = message
        self.date = date


class TemplateBase:
    """Provides a mechanism to map parameters (types or literals) to
    instantiations, following C++ mechanics.
    """

    def __init__(self, name, allow_default=True, scope=None):
        """
        Args:
            name: Name of the template object.
            allow_default: Allow a default value (None) to resolve to the
                parameters of the instantiation that was first added.
            scope: Parent scope for the template object.
        """
        self.name = name
        self.param_list = []
        self._allow_default = allow_default
        self._instantiation_map = {}
        self._instantiation_alias_map = {}
        if scope is None:
            scope = _get_module_from_stack()
        self._scope = scope
        self._instantiation_func = None
        self._deprecation_map = {}
        self.__doc__ = ""

    def __getitem__(self, *param):
        """Gets concrete class associate with the given arguments.

        Can be one of the following forms:
            template[param0]
            template[param0, param1, ...]
            template[(param0, param1, ...)]
            template[[param0, param1, ...]]
            template[None]   (first instantiation, if `allow_default` is True)
        """
        # For compatibility with Python3.
        if len(param) == 1:
            param = param[0]
        return self.get_instantiation(param)[0]

    def __call__(self, *args, **kwargs):
        """Permits pseudo argument deduction for a template. This is intended
        for use with pybind11, where any invalid argument raises a TypeError.

        This must have arguments passed to it.

        Note:
            While pybind11 has two passes for checking overloads (once
            without and once with conversions), this only passes through the
            instantiations once with conversions; therefore, be careful in how
            arguments are assigned.
        """
        if len(args) == 0 and len(kwargs) == 0:
            raise TypeError(
                f"{self.name}: incompatible function arguments for template: "
                "Cannot call without arguments"
            )
        result = self._call_internal(*args, **kwargs)
        if result is not None:
            return result
        raise TypeError(
            f"{self.name}: incompatible function arguments for template: "
            "No compatible instantiations"
        )

    def _call_internal(self, *args, **kwargs):
        """The implementation of __call__, to allow easy overriding."""
        for param in self.param_list:
            instantiation = self._instantiation_map[param]
            try:
                return instantiation(*args, **kwargs)
            except TypeError as e:
                if not _is_pybind11_type_error(e):
                    raise e

    def get_module_name(self):
        """
        Returns module name for this object's parent scope.

        Example:
            ::

                >>> pydrake.common.value.Value.get_module_name()
                pydrake.common.value
        """
        if isinstance(self._scope, types.ModuleType):
            return self._scope.__name__
        else:
            raise RuntimeError(
                f"Unable to resolve `get_module_name` for a scope that is not "
                f"a module: {self._scope}"
            )

    # Unique token to signify that this instantiation is deferred when using
    # `add_instantiations` or `define`. The instantiation function will not be
    # called until the specific instantiation is requested. To illustrate, the
    # following example defines a template via `@define`, and refers to
    # itself:
    #
    # @TemplateClass.define("MyTemplate", ((int,), (float,)))
    # def MyTemplate(param):
    #     print(MyTemplate)
    #     # ... Make and return a class.
    #
    # If the instantiation were not deferred and the inner function was called
    # before the decorator returned, an error would be raised since
    # `MyTemplate` is not yet defined. However, since it is deferred, this
    # should print out that `MyTemplate` is `<TemplateClass ...MyTemplate>`,
    # would only be called when the user requests something such as
    # `MyTemplate[float]`.
    class _Deferred:
        pass

    _deferred = _Deferred()

    def get_instantiation(self, param=None, throw_error=True):
        """Gets the instantiation for the given parameters.

        Args:
            param: Can be None, a single parameter, or a tuple/list of
                parameters.

        Returns:
            (instantiation, param), where `param` is the resolved set of
            parameters.
        """
        param = self._param_resolve(param)
        instantiation = self._instantiation_map.get(param)
        if instantiation is TemplateBase._deferred:
            assert self._instantiation_func is not None
            instantiation = self._instantiation_func(param)
            self._add_instantiation_internal(
                param, instantiation, skip_rename=False
            )
        elif instantiation is None and throw_error:
            raise RuntimeError(
                f"Invalid instantiation: {self._instantiation_name(param)}"
            )
        deprecation = self._deprecation_map.get(param)
        if deprecation is not None:
            _warn_deprecated(deprecation.message, date=deprecation.date)
        return (instantiation, param)

    def add_instantiation(self, param, instantiation, skip_rename=False):
        """Adds a unique instantiation.

        Note:
            `param` must not have already been added.
        """
        # Ensure that we do not already have this tuple.
        param = get_param_canonical(self._param_resolve(param))
        if param in self._instantiation_map:
            raise RuntimeError(
                f"Parameter instantiation already registered: {param}"
            )
        # Register it.
        self.param_list.append(param)
        self._add_instantiation_internal(param, instantiation, skip_rename)
        return param

    def _add_instantiation_internal(self, param, instantiation, skip_rename):
        # Adds instantiation. Permits overwriting for deferred cases.
        assert instantiation is not None
        if instantiation is not TemplateBase._deferred:
            old = instantiation
            instantiation = self._on_add(param, instantiation, skip_rename)
            assert instantiation is not None, (self, param, old)
            if instantiation is not old:
                self._instantiation_alias_map[old] = instantiation
        self._instantiation_map[param] = instantiation

    def add_instantiations(self, instantiation_func, param_list):
        """Adds a set of instantiations given a function and a list of
        parameter sets.

        Note:
            This method can only be called once.

        Args:
            instantiation_func: Function of the form `f(template, param)`,
                where `template` is the current template and `param` is the
                parameter set for the current instantiation.
            param_list: Ordered container of parameter sets that these
            instantiations should be produced for.
        """
        assert instantiation_func is not None
        if self._instantiation_func is not None:
            raise RuntimeError(
                "`add_instantiations` cannot be called multiple times."
            )
        self._instantiation_func = instantiation_func
        for param in param_list:
            self.add_instantiation(param, TemplateBase._deferred)

    def deprecate_instantiation(self, param, message, *, date=None):
        """Deprecates an instantiation for the given set of parameters.

        Note:
            This method can only be called once for a given instantiation.

        Args:
            param: Parameters for an instantiation that is already registered.
            message: Message to be shown when issuing a deprecation warning.
            date: (Optional) String of the form "YYYY-MM-DD".
                If supplied, will reformat the message to add the date as is
                done with DRAKE_DEPRECATED and its processing in mkdoc.py. This
                must be present if ``message`` does not contain the date
                itself.
        Returns:
            (instantiation, param), where ``param`` is the resolved parameters.
        """
        param = get_param_canonical(self._param_resolve(param))
        if param in self._deprecation_map:
            raise RuntimeError(
                f"Deprecation already registered: "
                f"{self._instantiation_name(param)}"
            )
        instantiation, param = self.get_instantiation(param)
        self._deprecation_map[param] = _Deprecation(message=message, date=date)
        return (instantiation, param)

    def get_param_set(self, instantiation):
        """Returns all parameters for a given `instantiation`.

        Returns:
            A set of instantiations.
        """
        param_list = []
        for param, check in self._instantiation_map.items():
            if check == instantiation:
                param_list.append(param)
        return set(param_list)

    def is_instantiation(self, obj):
        """Determines if an object is an instantion of the given template."""
        # Use `get_instantiation` so that we can handled deferred cases.
        for param in self.param_list:
            instantiation, _ = self.get_instantiation(param)
            obj = self._instantiation_alias_map.get(obj, obj)
            if instantiation is obj:
                return True
        return False

    def _param_resolve(self, param):
        # Resolves to canonical parameters, including default case.
        if param is None:
            assert self._allow_default
            assert len(self.param_list) > 0
            param = self.param_list[0]
        elif not isinstance(param, tuple):
            if not isinstance(param, list):
                # Assume scalar.
                param = (param,)
            else:
                param = tuple(param)
        return get_param_canonical(param)

    def _instantiation_name(self, param, *, mangle=False):
        """When mangle=False (the common case), returns the human-readable
        display name for an instantiation of the given ``param``s. This is
        typically a Python expression like ``LeafSystem_[AutoDiffXd]``.

        When mangle=True, returns the mangled name for an instantition, i.e.,
        a valid Python identifier that will be used as __name__ of the class.
        """
        if _is_building_documentation():
            # When we're building the website, the API docs should always use
            # the unmangled names. TODO(jwnimmer-tri) Ideally, we should have
            # our Sphinx extension demangle the names (so that we can remove
            # this giant hack), but at the moment this is the safest practical
            # way to ensure our website API reference displays correctly.
            mangle = False
        names = get_param_names(self._param_resolve(param), mangle=mangle)
        result = "{}[{}]".format(self.name, ",".join(names))
        if mangle:
            result = _MangledName.mangle(result)
        return result

    def _full_name(self):
        return "{}.{}".format(self._scope.__name__, self.name)

    def __str__(self):
        cls_name = pretty_class_name(type(self))
        return "<{} {}>".format(cls_name, self._full_name())

    def _on_add(self, param, instantiation, skip_rename):
        # To be overridden by child classes.
        return instantiation

    @classmethod
    def define(cls, name, param_list, *args, scope=None, **kwargs):
        """Provides a decorator for functions that defines a template using
        `name`. The template instantiations are added using
        `add_instantiations`, where the instantiation function is the decorated
        function.

        Args:
            name: Name of the template. This should generally match the name
                of the object being decorated for clarity.
            param_list: Ordered container of parameter sets. For more
                information, see `add_instantiations`.

        Note:
            The name of the inner class will not matter as it will be
            overwritten with the template instantiation name. In the below
            example, `Impl` will be renamed to `MyTemplate[int]` when
            `param=(int,)`.

        Example:
            ::

                @TemplateClass.define("MyTemplate",
                                      param_list=[(int,), (float,)])
                def MyTemplate(param):
                    T, = param
                    class Impl:
                        def __init__(self):
                            self.T = T
                    return Impl
        """
        if scope is None:
            scope = _get_module_from_stack()
        template = cls(name, *args, scope=scope, **kwargs)

        def decorator(instantiation_func):
            template.add_instantiations(instantiation_func, param_list)
            return template

        return decorator


class TemplateClass(TemplateBase):
    """Extension of `TemplateBase` for classes."""

    def __init__(self, name, *, scope=None, **kwargs):
        if scope is None:
            scope = _get_module_from_stack()
        TemplateBase.__init__(self, name, scope=scope, **kwargs)

    def _on_add(self, param, cls, skip_rename):
        # Unless this class was a default template instantiation, we need to
        # rename it now to describe its template arguments. (Most templated
        # C++ classes are bound using the TemporaryClassName() function.)
        if not skip_rename:
            cls._original_name = cls.__name__
            cls._original_qualname = getattr(cls, "__qualname__", cls.__name__)
            cls.__name__ = self._instantiation_name(param, mangle=True)
            # Define `__qualname__` in Python2 because that's what `pybind11`
            # uses when showing function signatures when an overload cannot be
            # found.
            # TODO(eric.cousineau): When porting to Python3 / six, try to
            # ensure this handles nesting.
            cls.__qualname__ = cls.__name__
            cls.__module__ = self._scope.__name__
            # Ensure instantiation is available for pickling... magically...
            setattr(self._scope, cls.__name__, cls)
        return cls

    def is_subclass_of_instantiation(self, obj):
        """Determines if `obj` is a subclass of one of the instantiations.

        Returns:
            The first instantiation of which `obj` is a subclass.
        """
        for param in self.param_list:
            instantiation, _ = self.get_instantiation(param)
            if issubclass(obj, instantiation):
                return instantiation
        return None


def _rename_callable(f, scope, name, cls=None):
    if isinstance(scope, type):
        module = scope.__module__
    else:
        module = scope.__name__
    # Renames a function.
    if (f.__module__, f.__name__) == (module, name):
        # Short circuit.
        return f
    if cls is not None:
        qualname = cls.__qualname__ + "." + name
    else:
        qualname = name
    # If Python2, we have to wrap instancemethods + built-in functions to spoof
    # the metadata.
    type_requires_wrap = (
        types.MethodType,
        types.BuiltinMethodType,
        types.BuiltinFunctionType,
    )
    if isinstance(f, type_requires_wrap):
        orig = f

        def f(*args, **kwargs):
            return orig(*args, **kwargs)

        f.__module__ = module
        f.__name__ = name
        f.__qualname__ = qualname
        f.__doc__ = orig.__doc__
    else:
        f.__module__ = module
        f.__name__ = name
        f.__qualname__ = qualname
    return f


class TemplateFunction(TemplateBase):
    """Extension of `TemplateBase` for functions."""

    def _on_add(self, param, func, skip_rename):
        assert skip_rename is False
        new_name = self._instantiation_name(param, mangle=True)
        func = _rename_callable(func, self._scope, new_name)
        setattr(self._scope, func.__name__, func)
        return func


class TemplateMethod(TemplateBase):
    """Extension of `TemplateBase` for class methods."""

    def __init__(self, name, cls, scope=None, **kwargs):
        if scope is None:
            scope = _get_module_from_stack()
        TemplateBase.__init__(self, name, scope=scope, **kwargs)
        # TODO(eric.cousineau): Merge `cls` into `scope` once we are Python 3
        # only.
        self._cls = cls

    def _on_add(self, param, func, skip_rename):
        assert skip_rename is False
        new_name = self._instantiation_name(param, mangle=True)
        func = _rename_callable(func, self._scope, new_name, self._cls)
        setattr(self._cls, func.__name__, func)
        return func

    def __get__(self, obj, objtype):
        """Provides descriptor accessor."""
        if obj is None:
            return self
        else:
            return TemplateMethod._Bound(self, obj)

    def __set__(self, obj, value):
        raise RuntimeError("Read-only property")

    def __str__(self):
        return f"<unbound TemplateMethod {self._full_name()}>"

    def _full_name(self):
        return "{}.{}".format(self._cls.__name__, self.name)

    class _Bound:
        def __init__(self, template, obj):
            self._tpl = template
            self._obj = obj

        def __getitem__(self, param):
            unbound = self._tpl[param]
            bound = types.MethodType(unbound, self._obj)
            return bound

        def __str__(self):
            return "<bound TemplateMethod {} of {}>".format(
                self._tpl._full_name(), self._obj
            )
